import pandas as pd
import numpy as np
import pymc3 as pm
from pymc3.model import FreeRV
import matplotlib.pyplot as plt
from bambi.external.six import string_types
from collections import OrderedDict, defaultdict
from bambi.utils import listify
from patsy import dmatrices, dmatrix
import re, warnings
from bambi.priors import PriorFactory, PriorScaler, Prior
from copy import deepcopy
import statsmodels.api as sm


class Model(object):

    '''
    Args:
        data (DataFrame, str): the dataset to use. Either a pandas
            DataFrame, or the name of the file containing the data, which
            will be passed to pd.read_table().
        intercept (bool): If True, an intercept term is added to the model
            at initialization. Defaults to False, as both fixed and random
            effect specifications will add an intercept by default.
        backend (str): The name of the BackEnd to use. Currently only
            'pymc3' is supported.
        default_priors (dict, str): An optional specification of the
            default priors to use for all model terms. Either a dict
            containing named distributions, families, and terms (see the
            documentation in priors.PriorFactory for details), or the name
            of a JSON file containing the same information.
        auto_scale (bool): If True (default), priors are automatically rescaled
            to the data (to be weakly informative) any time default priors are
            used. Note that any priors explicitly set by the user will always
            take precedence over default priors.
        dropna (bool): When True, rows with any missing values in either the
            predictors or outcome are automatically dropped from the dataset in
            a listwise manner.
        taylor (int): Order of Taylor expansion to use in approximate variance
            when constructing the default priors. Should be between 1 and 13.
            Lower values are less accurate, tending to undershoot the correct
            prior width, but are faster to compute and more stable. Odd-
            numbered values tend to work better. Defaults to 5 for Normal 
            models and 1 for non-Normal models. Values higher than the defaults
            are generally not recommended as they can be unstable.
    '''

    def __init__(self, data=None, intercept=False, backend='pymc3',
                 default_priors=None, auto_scale=True, dropna=False,
                 taylor=None):

        if isinstance(data, string_types):
            data = pd.read_table(data, sep=None)

        self.default_priors = PriorFactory(default_priors)

        obj_cols = data.select_dtypes(['object']).columns
        data[obj_cols] = data[obj_cols].apply(lambda x: x.astype('category'))
        self.data = data
        # Some random effects stuff later requires us to make guesses about
        # column groupings into terms based on patsy's naming scheme.
        if re.search("[\[\]]+", ''.join(data.columns)):
            warnings.warn("At least one of the column names in the specified "
                          "dataset contain square brackets ('[' or ']')."
                          "This may cause unexpected behavior if you specify "
                          "models with random effects. You are encouraged to "
                          "rename your columns to avoid square brackets.")
        self.reset()

        if backend.lower() == 'pymc3':
            from bambi.backends import PyMC3BackEnd
            self.backend = PyMC3BackEnd()
        else:
            raise ValueError(
                "At the moment, only the PyMC3 backend is supported.")

        if intercept:
            self.add_intercept()

        self.auto_scale = auto_scale
        self.dropna = dropna
        self.taylor = taylor

    def reset(self):
        '''
        Reset list of terms and y-variable.
        '''
        self.terms = OrderedDict()
        self.y = None

    def build(self):
        ''' Set up the model for sampling/fitting. Performs any steps that
        require access to all model terms (e.g., scaling priors on each term),
        then calls the BackEnd's build() method.
        '''
        if self.y is None:
            raise ValueError("No outcome (y) variable is set! Please call "
                             "add_y() or specify an outcome variable using the"
                             " formula interface before build() or fit().")

        # Check for NaNs and halt if dropna is False--otherwise issue warning.
        arrs = []
        for t in self.terms.values():
            if isinstance(t.data, dict):
                arrs.extend(list(t.data.values()))
            else:
                arrs.append(t.data)
        X = np.concatenate(arrs + [self.y.data], axis=1)
        na_index = np.isnan(X).any(1)
        if na_index.sum():
            msg = "%d rows were found contain at least one missing value." \
                % na_index.sum()
            if not self.dropna:
                msg += "Please make sure the dataset contains no missing " \
                       "values. Alternatively, if you want rows with missing " \
                       "values to be automatically deleted in a list-wise " \
                       "manner (not recommended), please set dropna=True at " \
                       "model initialization."
                raise ValueError(msg)

            # warn and then remove missing values
            msg += " Automatically removing %d rows from the dataset." \
                % na_index.sum()
            warnings.warn(msg)
            keeps = np.invert(na_index)
            for t in self.fixed_terms.values():
                t.data = t.data[keeps]
            self.y.data = self.y.data[keeps]

        # X = fixed effects design matrix (excluding intercept/constant term)
        # r2_x = 1 - 1/VIF, i.e., R2 for predicting each x from all other x's.
        # only compute these stats if there are multiple terms in the model
        terms = [t for t in self.fixed_terms.values() if t.name != 'Intercept']

        if len(self.fixed_terms) > 1:

            X = [pd.DataFrame(x.data, columns=x.levels) for x in terms]
            X = pd.concat(X, axis=1)

            # interim solution for handling non-normal models
            sd_y_defaults = {
                'gaussian': {
                    'identity': self.y.data.std(),
                    'logit': self.y.data.std(),
                    'probit': self.y.data.std(),
                    'inverse': self.y.data.std(),
                    'log': self.y.data.std()
                },
                'binomial': {
                    'identity': self.y.data.std(),
                    'logit': np.pi / 3**.5,
                    'probit': 1,
                    'inverse': self.y.data.std(),
                    'log': self.y.data.std()
                },
                'poisson': {
                    'identity': self.y.data.std(),
                    'logit': self.y.data.std(),
                    'probit': self.y.data.std(),
                    'inverse': self.y.data.std(),
                    'log': self.y.data.std()
                },
                't': {
                    'identity': self.y.data.std(),
                    'logit': self.y.data.std(),
                    'probit': self.y.data.std(),
                    'inverse': self.y.data.std(),
                    'log': self.y.data.std()
                }
            }

            self.dm_statistics = {
                'r2_x': pd.Series({
                    x: sm.OLS(endog=X[x],
                        exog=sm.add_constant(X.drop(x, axis=1)) \
                            if 'Intercept' in self.term_names \
                            else X.drop(x, axis=1)).fit().rsquared
                    for x in list(X.columns)}),
                'sd_x': X.std(),
                'sd_y': sd_y_defaults[self.family.name][self.family.link],
                'mean_x': X.mean(axis=0)
            }

            # save potentially useful info for diagnostics, send to ModelResults
            # mat = correlation matrix of X, w/ diagonal replaced by X means
            mat = X.corr()
            for x in list(mat.columns):
                mat.loc[x, x] = self.dm_statistics['mean_x'][x]
            self._diagnostics = {
                # the Variance Inflation Factors (VIF), which is possibly useful
                # for diagnostics
                'VIF': 1/(1 - self.dm_statistics['r2_x']),
                'corr_mean_X': mat
            }

            # throw informative error if perfect collinearity among fixed fx
            if any(self.dm_statistics['r2_x'] > .999):
                raise ValueError(
                    "There is perfect collinearity among the fixed effects!\n"+\
                    "Printing some design matrix statistics:\n" + \
                    str(self.dm_statistics) + '\n' + \
                    str(self._diagnostics))

        # throw informative error message if any categorical predictors have 1 category
        if any(np.array([x.data.size for x in self.fixed_terms.values()])==0):
            raise ValueError("At least one categorical predictor contains only 1 category!")

        # only set priors if there is at least one term in the model
        if len(self.terms) > 0:
            # Get and scale default priors if none are defined yet
            if self.taylor is not None:
                taylor = self.taylor
            else:
                taylor = 5 if self.family.name=='gaussian' else 1
            scaler = PriorScaler(self, taylor=taylor)
            scaler.scale()

        # For binomial models with n_trials = 1 (most common use case),
        # tell user which event is being modeled
        if self.family.name=='binomial' and np.max(self.y.data) < 1.01:
            event = next(i for i,x in enumerate(self.y.data.flatten()) if x>.99)
            warnings.warn('Modeling the probability that {}==\'{}\''.format(
                self.y.name, str(self.data[self.y.name].iloc[event])))

        self.backend.build(self)
        self.built = True

    def fit(self, fixed=None, random=None, priors=None, family='gaussian',
            link=None, run=True, categorical=None, **kwargs):
        '''
        Fit the model using the current BackEnd.
        Args:
            fixed (str): Optional formula specification of fixed effects.
            random (list): Optional list-based specification of random effects.
            priors (dict): Optional specification of priors for one or more
                terms. A dict where the keys are the names of terms in the
                model, and the values are either instances of class Prior or
                ints, floats, or strings that specify the width of the priors
                on a standardized scale.
            family (str, Family): A specification of the model family
                (analogous to the family object in R). Either a string, or an
                instance of class priors.Family. If a string is passed, a
                family with the corresponding name must be defined in the
                defaults loaded at Model initialization. Valid pre-defined
                families are 'gaussian', 'binomial', 'poisson', and 't'.
            link (str): The model link function to use. Can be either a string
                (must be one of the options defined in the current backend;
                typically this will include at least 'identity', 'logit',
                'inverse', and 'log'), or a callable that takes a 1D ndarray
                or theano tensor as the sole argument and returns one with
                the same shape.
            run (bool): Whether or not to immediately begin fitting the model
                once any set up of passed arguments is complete.
            categorical (str, list): The names of any variables to treat as
                categorical. Can be either a single variable name, or a list
                of names. If categorical is None, the data type of the columns
                in the DataFrame will be used to infer handling. In cases where
                numeric columns are to be treated as categoricals (e.g., random
                factors coded as numerical IDs), explicitly passing variable
                names via this argument is recommended.
        '''
        if fixed is not None or random is not None:
            self.add_formula(fixed=fixed, random=random, priors=priors,
                             family=family, link=link, categorical=categorical,
                             append=False)
        ''' Run the BackEnd to fit the model. '''
        if run:
            if not self.built:
                warnings.warn("Current Bayesian model has not been built yet; "
                              "building it first before sampling begins.")
                self.build()
            return self.backend.run(**kwargs)

    def add_intercept(self):
        '''
        Adds a constant term to the model. Generally unnecessary when using the
        formula interface, but useful when specifying the model via add_term().
        '''
        n = len(self.data)
        df = pd.DataFrame(np.ones((n, 1)), columns=['Intercept'])
        self.add_term('Intercept', df)

    def add_formula(self, fixed=None, random=None, priors=None,
                    family='gaussian', link=None, categorical=None,
                    append=True):
        '''
        Adds one or more terms to the model via an R-like formula syntax.
        Args:
            fixed (str): Optional formula specification of fixed effects.
            random (list): Optional list-based specification of random effects.
            priors (dict): Optional specification of priors for one or more
                terms. A dict where the keys are the names of terms in the
                model, and the values are either instances of class Prior or
                ints, floats, or strings that specify the width of the priors
                on a standardized scale.
            family (str, Family): A specification of the model family
                (analogous to the family object in R). Either a string, or an
                instance of class priors.Family. If a string is passed, a
                family with the corresponding name must be defined in the
                defaults loaded at Model initialization. Valid pre-defined
                families are 'gaussian', 'binomial', 'poisson', and 't'.
            link (str): The model link function to use. Can be either a string
                (must be one of the options defined in the current backend;
                typically this will include at least 'identity', 'logit',
                'inverse', and 'log'), or a callable that takes a 1D ndarray
                or theano tensor as the sole argument and returns one with
                the same shape.
            categorical (str, list): The names of any variables to treat as
                categorical. Can be either a single variable name, or a list
                of names. If categorical is None, the data type of the columns
                in the DataFrame will be used to infer handling. In cases where
                numeric columns are to be treated as categoricals (e.g., random
                factors coded as numerical IDs), explicitly passing variable
                names via this argument is recommended.
            append (bool): if True, terms are appended to the existing model
                rather than replacing any existing terms. This allows
                formula-based specification of the model in stages.
        '''
        data = self.data

        if priors is None:
            priors = {}

        if not append:
            self.reset()

        if fixed is not None:
            # Explicitly convert columns to category if desired--though this
            # can also be done within the formula using C().
            if categorical is not None:
                data = data.copy()
                cats = listify(categorical)
                data[cats] = data[cats].apply(lambda x: x.astype('category'))

            if '~' in fixed:
                # check to see if formula is using the 'y[event] ~ x' syntax
                # (for binomial models). If so, chop it into groups:
                # 1 = 'y[event]', 2 = 'y', 3 = 'event', 4 = 'x'
                # If this syntax is not being used, event = None
                event = re.match(r'^((\S+)\[(\S+)\])\s*~(.*)$', fixed)
                if event is not None:
                    fixed = '{}~{}'.format(event.group(2),event.group(4))
                y, X = dmatrices(fixed, data=data)
                y_label = y.design_info.term_names[0]
                if event is not None:
                    # pass in new Y data that has 1 if y=event and 0 otherwise
                    y_data = y[:, 
                               y.design_info.column_names.index(event.group(1))]
                    y_data = pd.DataFrame({event.group(3): y_data})
                    self.add_y(y_label, family=family, link=link, data=y_data)
                else:
                    # use Y as-is
                    self.add_y(y_label, family=family, link=link)
            else:
                X = dmatrix(fixed, data=data)

            # Loop over predictor terms
            for _name, _slice in X.design_info.term_name_slices.items():
                cols = X.design_info.column_names[_slice]
                term_data = pd.DataFrame(X[:, _slice], columns=cols)
                prior = priors.pop(_name, priors.pop('fixed', None))
                self.add_term(_name, data=term_data, prior=prior)

        # Random effects
        if random is not None:
            random = listify(random)
            for f in random:
                f = f.strip()
                kwargs = {'random': True}
                if re.search('[\*\(\)]+', f):
                    raise ValueError("Random term '%s' contains an invalid "
                                     "character. Note that only the | and + "
                                     "operators are currently supported in "
                                     "random effects specifications.")

                # replace explicit intercept terms like '1|subj' with 'subj'
                f = re.sub(r'^1\s*\|(.*)', r'\1', f).strip()

                # Split specification into intercept, predictor, and grouper
                patt = r'^([01]+)*[\s\+]*([^\|]+)\|*(.*)'
                intcpt, pred, grpr = re.search(patt, f).groups()
                label = '{}|{}'.format(pred, grpr) if grpr else pred

                # Default to including random intercepts
                if intcpt is None:
                    intcpt = 1
                intcpt = int(intcpt)

                # If there's no grouper, we must be adding random intercepts
                if not grpr:
                    kwargs.update(dict(categorical=True, drop_first=False))
                    variable = pred

                else:
                    # If we're adding slopes, add random intercepts as well,
                    # unless they were explicitly excluded
                    if intcpt and grpr not in self.terms:
                        self.add_term(variable=grpr, categorical=True,
                                      random=True, drop_first=False)
                    if self.data[pred].dtype.name in ['object', 'category']:
                        kwargs['categorical'] = True
                        if not intcpt:
                            kwargs['drop_first'] = False
                    variable, kwargs['over'] = pred, grpr

                prior = priors.pop(label, priors.pop('random', None))
                self.add_term(variable=variable, label=label, **kwargs)

    def add_y(self, variable, prior=None, family='gaussian', link=None, *args,
              **kwargs):
        '''
        Add a dependent (or outcome) variable to the model.
        Args:
            variable (str): the name of the dataset column containing the
                y values.
            prior (Prior, int, float, str): Optional specification of prior.
                Can be an instance of class Prior, a numeric value, or a string
                describing the width. In the numeric case, the distribution
                specified in the defaults will be used, and the passed value
                will be used to scale the appropriate variance parameter. For
                strings (e.g., 'wide', 'narrow', 'medium', or 'superwide'),
                predefined values will be used.
            family (str, Family): A specification of the model family
                (analogous to the family object in R). Either a string, or an
                instance of class priors.Family. If a string is passed, a
                family with the corresponding name must be defined in the
                defaults loaded at Model initialization. Valid pre-defined
                families are 'gaussian', 'binomial', 'poisson', and 't'.
            link (str): The model link function to use. Can be either a string
                (must be one of the options defined in the current backend;
                typically this will include at least 'identity', 'logit',
                'inverse', and 'log'), or a callable that takes a 1D ndarray
                or theano tensor as the sole argument and returns one with
                the same shape.
            args, kwargs: Optional positional and keyword arguments to pass
                onto add_term().
        '''
        if isinstance(family, string_types):
            family = self.default_priors.get(family=family)
        self.family = family

        # Override family's link if another is explicitly passed
        if link is not None:
            self.family.link = link

        if prior is None:
            prior = self.family.prior

        # implement default Uniform [0, sd(Y)] prior for residual SD
        if self.family.name == 'gaussian':
            prior.update(sd=Prior('Uniform', lower=0,
                upper=self.data[variable].std()))

        self.add_term(variable, prior=prior, *args, **kwargs)
        # use last-added term name b/c it could have been changed by add_term
        name = list(self.terms.values())[-1].name
        self.y = self.terms.pop(name)
        self.built = False

    def add_term(self, variable, data=None, label=None, categorical=False,
                 random=False, over=None, prior=None, drop_first=True):
        '''
        Add a term to the model.
        Args:
            variable (str): The name of the dataset column to use; also used
                as the Term instance label if not otherwise specified using
                the label argument.
            data (DataFrame): Optional pandas DataFrame containing the term
                values to use. If None (default), the correct column will be
                extracted from the dataset currently loaded into the model
                (based on the name passed in the variable argument).
            label (str): Optional label/name to use for the term. If None, the
                label will be automatically generated based on the variable
                name and additional arguments.
            categorical (bool): Whether or not the input variable should be
                treated as categorical (defaults to False).
            random (bool): If True, the predictor variable is modeled as a
                random effect; if False, the predictor is modeled as a fixed
                effect.
            over (str): When adding random slopes, the name of the variable the
                slopes are randomly distributed over. For example, if
                variable='condition', categorical=True, random=True, and
                over='subject', a separate set of random subject slopes will be
                added for each level of the condition variable. This is
                analogous to the lme4 specification of 'condition|subject'.
            prior (Prior, int, float, str): Optional specification of prior.
                Can be an instance of class Prior, a numeric value, or a string
                describing the width. In the numeric case, the distribution
                specified in the defaults will be used, and the passed value
                will be used to scale the appropriate variance parameter. For
                strings (e.g., 'wide', 'narrow', 'medium', or 'superwide'),
                predefined values will be used.
            drop_first (bool): indicates whether to use full rank or N-1 coding
                when the predictor is categorical. If True, the N levels of the
                categorical variable will be represented using N dummy columns.
                If False, the predictor will be represented using N-1 binary
                indicators, where each indicator codes the contrast between
                the N_j and N_0 columns, for j = {1..N-1}.

        Notes: One can think of bambi's split_by operation as a sequence of two
            steps. First, the target variable is multiplied by the splitting
            variable. This is equivalent to a formula call like 'A:B'. Second,
            the columns of the resulting matrix are "grouped" by the levels
            of the split_by variable.
        '''

        if data is None:
            data = self.data.copy()

        # Make sure user didn't forget to set categorical=True
        if variable in data.columns and \
                data.loc[:, variable].dtype.name in ['object', 'category']:
            categorical = True

        else:
            # If all columns have identical names except for levels in [],
            # assume they've already been contrast-coded, and pass data as-is
            cols = [re.sub('\[.*?\]', '', c) for c in data.columns]
            if len(set(cols)) > 1:
                X = data[[variable]]

        if categorical:
            X = pd.get_dummies(data[variable], drop_first=drop_first)
        elif variable in data.columns:
            X = data[[variable]]
        else:
            X = data

        if random and over is not None:
            id_var = pd.get_dummies(data[over], drop_first=False)
            data = {over: id_var.values, variable: X.values}
            f = '0 + %s:%s' % (over, variable)
            data = dmatrix(f, data=data)
            cols = data.design_info.column_names
            data = pd.DataFrame(data, columns=cols)

            # For categorical effects, one variance term per predictor level
            if categorical:
                split_data = {}
                groups = list(set([re.sub(r'^.*?\:', '', c) for c in cols]))
                for g in groups:
                    patt = re.escape(r':%s' % g) + '$'
                    level_data = data.filter(regex=patt)
                    level_data.columns = [
                        c.split(':')[0] for c in level_data.columns]
                    level_data = level_data.loc[
                        :, (level_data != 0).any(axis=0)]
                    split_data[g] = level_data.values
                data = split_data
        else:
            data = X

        if label is None:
            label = variable
            if over is not None:
                label += '|%s' % over

        term = Term(name=label, data=data, categorical=categorical,
                    random=random, prior=prior)
        self.terms[term.name] = term
        self.built = False

    def set_priors(self, priors=None, fixed=None, random=None):
        '''
        Set priors for one or more existing terms.
        Args:
            priors (dict): Dict of priors to update. Keys are names of terms
                to update; values are the new priors (either a Prior instance,
                or an int or float that scales the default priors). Note that
                a tuple can be passed as the key, in which case the same prior
                will be applied to all terms named in the tuple.
            fixed (Prior, int, float, str): a prior specification to apply to
                all fixed terms currently included in the model.
            random (Prior, int, float, str): a prior specification to apply to
                all random terms currently included in the model.
        '''

        targets = {}

        if fixed is not None:
            targets.update({name: fixed for name in self.fixed_terms.keys()})

        if random is not None:
            targets.update({name: random for name in self.random_terms.keys()})

        if priors is not None:
            for k, prior in priors.items():
                for name in listify(k):
                    if name not in self.terms:
                        raise ValueError("The model contains no term with "
                                         "the name '%s'." % name)
                    targets[name] = prior

        for name, prior in targets.items():
            self.terms[name].prior = prior

    def plot(self, kind='priors'):
        # Currently this only supports plotting priors for fixed effects
        if not self.built:
            raise ValueError("Cannot plot priors until model is built!")

        with pm.Model():
            # get priors for fixed fx, separately for each level of each predictor
            dists = []
            for t in self.fixed_terms.values():
                for i,l in enumerate(t.levels):
                    params = {k: v[i % len(v)] \
                        if isinstance(v, np.ndarray) else v
                        for k,v in t.prior.args.items()}
                    dists += [getattr(pm, t.prior.name)(l, **params)]

            # get priors for random effect SDs
            for t in self.random_terms.values():
                prior = t.prior.args['sd'].name
                params = t.prior.args['sd'].args
                dists += [getattr(pm, prior)(t.name+'_sd', **params)]

            # add priors on Y params if applicable
            y_prior = [(k,v) for k,v in self.y.prior.args.items()
                if isinstance(v, Prior)]
            if len(y_prior):
                for p in y_prior:
                    dists += [getattr(pm, p[1].name)('_'.join([self.y.name,
                        p[0]]), **p[1].args)]
            
            # make the plot!
            p = float(len(dists))
            fig, axes = plt.subplots(int(np.ceil(p/2)), 2,
                figsize=(12,np.ceil(p/2)*2))
            # in case there is only 1 row
            if int(np.ceil(p/2))<2: axes = axes[None,:]
            for i,d in enumerate(dists):
                dist = d.distribution if isinstance(d, FreeRV) else d
                samp = pd.Series(dist.random(size=1000).flatten())
                samp.plot(kind='hist', ax=axes[divmod(i,2)[0], divmod(i,2)[1]],
                    normed=True)
                samp.plot(kind='kde', ax=axes[divmod(i,2)[0], divmod(i,2)[1]],
                    color='b')
                axes[divmod(i,2)[0], divmod(i,2)[1]].set_title(d.name)
            fig.tight_layout()
        
        return axes

    @property
    def term_names(self):
        ''' Return names of all terms in order of addition to model. '''
        return list(self.terms.keys())

    @property
    def fixed_terms(self):
        ''' Return dict of all and only fixed effects in model. '''
        return {k: v for (k, v) in self.terms.items() if not v.random}

    @property
    def random_terms(self):
        ''' Return dict of all and only random effects in model. '''
        return {k: v for (k, v) in self.terms.items() if v.random}


class Term(object):

    '''
    Representation of a single model term.
    Args:
        name (str): Name of the term.
        data (DataFrame, Series, ndarray): The term values.
        categorical (bool): If True, the source variable is interpreted as
            nominal/categorical. If False, the source variable is treated
            as continuous.
        prior (Prior): A specification of the prior(s) to use. An instance
            of class priors.Prior.
    '''
    def __init__(self, name, data, categorical=False, random=False, prior=None):

        self.name = name
        self.categorical = categorical
        self.random = random
        self.prior = prior

        if isinstance(data, pd.Series):
            data = data.to_frame()
        if isinstance(data, pd.DataFrame):
            self.levels = list(data.columns)
            data = data.values
        elif isinstance(data, dict):
            pass   # Random effects pass through here
        else:
            data = np.atleast_2d(data)
            self.levels = list(range(data.shape[1]))

        self.data = data



import numpy as np
import pandas as pd
from scipy.special import hyp2f1
from pandas import Series
from os.path import dirname, join
from bambi.external.six import string_types
from copy import deepcopy
import json
import re
import statsmodels.api as sm


class Family(object):

    '''
    A specification of model family.
    Args:
        name (str): Family name
        prior (Prior): A Prior instance specifying the model likelihood prior
        link (str): The name of the link function transforming the linear
            model prediction to a parameter of the likelihood
        parent (str): The name of the prior parameter to set to the link-
            transformed predicted outcome (e.g., mu, p, etc.).
    '''

    def __init__(self, name, prior, link, parent):
        self.name = name
        self.prior = prior
        self.link = link
        self.parent = parent
        fams = {
            'gaussian': sm.families.Gaussian,
            'binomial': sm.families.Binomial,
            'poisson': sm.families.Poisson,
            't': None # not implemented in statsmodels
        }
        self.smfamily = fams[name] if name in fams.keys() else None


class Prior(object):

    '''
    Abstract specification of a term prior.
    Args:
        name (str): Name of prior distribution (e.g., Normal, Binomial, etc.)
        kwargs (dict): Optional keywords specifying the parameters of the
            named distribution.
    '''

    def __init__(self, name, **kwargs):
        self.name = name
        self.args = {}
        self.update(**kwargs)

    def update(self, **kwargs):
        '''
        Update the model arguments with additional arguments.
        Args:
            kwargs (dict): Optional keyword arguments to add to prior args.
        '''
        self.args.update(kwargs)


class PriorFactory(object):

    '''
    An object that supports specification and easy retrieval of default priors.
    Args:
        defaults (str, dict): Optional base configuration containing default
            priors for distribution, families, and term types. If a string,
            the name of a JSON file containing the config. If a dict, must
            contain keys for 'dists', 'terms', and 'families'; see the built-in
            JSON configuration for an example. If None, a built-in set of
            priors will be used as defaults.
        dists (dict): Optional specification of named distributions to use
            as priors. Each key gives the name of a newly defined distribution;
            values are two-element lists, where the first element is the name
            of the built-in distribution to use ('Normal', 'Cauchy', etc.),
            and the second element is a dictionary of parameters on that
            distribution (e.g., {'mu': 0, 'sd': 10}). Priors can be nested
            to arbitrary depths by replacing any parameter with another prior
            specification.
        terms (dict): Optional specification of default priors for different
            model term types. Valid keys are 'intercept', 'fixed', or 'random'.
            Values are either strings preprended by a #, in which case they
            are interpreted as pointers to distributions named in the dists
            dictionary, or key -> value specifications in the same format as
            elements in the dists dictionary.
        families (dict): Optional specification of default priors for named
            family objects. Keys are family names, and values are dicts
            containing mandatory keys for 'dist', 'link', and 'parent'.

    Examples:
        >>> dists = { 'my_dist': ['Normal', {'mu': 10, 'sd': 1000}]}
        >>> pf = PriorFactory(dists=dists)

        >>> families = { 'normalish': { 'dist': ['normal', {sd: '#my_dist'}],
        >>>                             link:'identity', parent: 'mu'}}
        >>> pf = PriorFactory(dists=dists, families=families)
    '''

    def __init__(self, defaults=None, dists=None, terms=None, families=None):

        if defaults is None:
            defaults = join(dirname(__file__), 'config', 'priors.json')

        if isinstance(defaults, string_types):
            defaults = json.load(open(defaults, 'r'))

        # Just in case the user plans to use the same defaults elsewhere
        defaults = deepcopy(defaults)

        if isinstance(dists, dict):
            defaults['dists'].update(dists)

        if isinstance(terms, dict):
            defaults['terms'].update(terms)

        if isinstance(families, dict):
            defaults['families'].update(families)

        self.dists = defaults['dists']
        self.terms = defaults['terms']
        self.families = defaults['families']

    def _get_prior(self, spec):

        if isinstance(spec, string_types):
            spec = re.sub('^\#', '', spec)
            return self._get_prior(self.dists[spec])
        elif isinstance(spec, (list, tuple)):
            name, args = spec
            if name.startswith('#'):
                name = re.sub('^\#', '', name)
                prior = self._get_prior(self.dists[name])
            else:
                prior = Prior(name)
            args = {k: self._get_prior(v) for (k, v) in args.items()}
            prior.update(**args)
            return prior
        else:
            return spec

    def get(self, dist=None, term=None, family=None, **kwargs):
        '''
        Retrieve default prior for a named distribution, term type, or family.
        Args:
            dist (str): Name of desired distribution. Note that the name is
                the key in the defaults dictionary, not the name of the
                Distribution object used to construct the prior.
            term (str): The type of term family to retrieve defaults for.
                Must be one of 'intercept', 'fixed', or 'random'.
            family (str): The name of the Family to retrieve. Must be a value
                defined internally. In the default config, this is one of
                'gaussian', 'binomial', 'poisson', or 't'.
        '''
        if dist is not None:
            if dist not in self.dists:
                raise ValueError(
                    "'%s' is not a valid distribution name." % dist)
            return self._get_prior(self.dists[dist])
        elif term is not None:
            if term not in self.terms:
                raise ValueError("'%s' is not a valid term type." % term)
            return self._get_prior(self.terms[term])
        elif family is not None:
            if family not in self.families:
                raise ValueError("'%s' is not a valid family name." % family)
            _f = self.families[family]
            prior = self._get_prior(_f['dist'])
            return Family(family, prior, _f['link'], _f['parent'])


class PriorScaler(object):

    # Default is 'wide'. The wide prior SD is sqrt(1/3) = .577 on the partial
    # corr scale, which is the SD of a flat prior over [-1,1].
    names = {
        'narrow': 0.2,
        'medium': 0.4,
        'wide': 3 ** -0.5,
        'superwide': 0.8
    }

    def __init__(self, model, taylor):
        self.model = model
        self.stats = model.dm_statistics if hasattr(model, 'dm_statistics') \
            else None
        self.dm = pd.DataFrame({'{}[{}]'.format(t.name, lev): t.data[:, lev]
                   for t in model.fixed_terms.values()
                   for lev in range(len(t.levels))})
        self.priors = {}
        self.mle = sm.GLM(endog=self.model.y.data, exog=self.dm,
            family=self.model.family.smfamily(),
            missing='drop' if self.model.dropna else 'none').fit()
        self.taylor = taylor
        with open(join(dirname(__file__),'config','derivs.txt'), 'r') as file:
            self.deriv = [next(file).strip('\n') for x in range(taylor+1)]

    def _get_slope_stats(self, exog, predictor, sd_corr, full_mod=None,
        points=4):
        # full_mod: statsmodels GLM to replace MLE model. For when 'predictor'
        #     is not in the fixed part of the model.
        # points: number of points to use for LL approximation

        if full_mod is None:
            full_mod = self.mle

        # figure out which column of exog to drop for the null model
        keeps = [i for i,x in enumerate(list(exog.columns))
            if not np.array_equal(predictor, exog[x].values.flatten())]
        i = [x for x in range(exog.shape[1]) if x not in keeps][0]

        # get log-likelihood values from beta=0 to beta=MLE
        values = np.linspace(0., full_mod.params[i], points)
        increment = np.diff(values)[0]
        # if there are multiple predictors, use statsmodels to optimize the LL
        if keeps:
            null = [sm.GLM(endog=self.model.y.data, exog=exog,
                family=self.model.family.smfamily()).fit_constrained(
                str(exog.columns[i])+'='+str(val),
                start_params=full_mod.params.values)
                for val in values[:-1]]
            null = np.append(null, full_mod)
            ll = np.array([x.llf for x in null])
        # if just a single predictor, use statsmodels to evaluate the LL
        else:
            null = [sm.GLM(endog=self.model.y.data,
                exog=exog, family=self.model.family.smfamily()).loglike(
                np.squeeze(self.model.y.data), val*predictor)
                for val in values[:-1]]
            ll = np.append(null, full_mod.llf)

        # compute params of quartic approximatino to log-likelihood
        # c: intercept, d: shift parameter
        # a: quartic coefficient, b: quadratic coefficient
        c, d = ll[-1], -np.asscalar(full_mod.params[i])
        X = np.matrix([(values+d)**4,
                       (values+d)**2]).T
        a, b = np.squeeze((np.linalg.inv(X.T * X) * X.T * (ll[:,None] - c)).A)

        # m, v: mean and variance of beta distribution of correlations
        # p, q: corresponding shape parameters of beta distribution
        m = .5
        v = sd_corr**2/4
        p = m*(m*(1-m)/v - 1)
        q = (1-m)*(m*(1-m)/v - 1)

        # function to return central moments of rescaled beta distribution
        def moment(k): return (2*p/(p+q))**k * hyp2f1(p, -k, p+q, (p+q)/p)

        # evaluate the derivatives of beta = f(correlation).
        # dict 'point' gives points about which to Taylor expand. We want to 
        # expand about the mean (generally 0), but some of the derivatives
        # do not exist at 0. Evaluating at a point very close to 0 (e.g., .001)
        # generally gives good results, but the higher order the expansion, the
        # further from 0 we need to evaluate the derivatives, or they blow up.
        point = dict(zip(range(1,14), 2**np.linspace(-1,5,13)/100))
        vals = dict(a=a, b=b, n=len(self.model.y.data), r=point[self.taylor])
        _deriv = [eval(x, globals(), vals) for x in self.deriv]

        # compute and return the approximate SD
        def term(i, j):
            return 1/np.math.factorial(i) * 1/np.math.factorial(j) \
                * _deriv[i] * _deriv[j] \
                * (moment(i+j) - moment(i)*moment(j))
        terms = [term(i,j)
            for i in range(1,self.taylor+1) for j in range(1,self.taylor+1)]
        return np.array(terms).sum()**.5

    def _get_intercept_stats(self, add_slopes=True):
        # start with mean and variance of Y on the link scale
        mod = sm.GLM(endog=self.model.y.data,
            exog=np.repeat(1, len(self.model.y.data)),
            family=self.model.family.smfamily(),
            missing='drop' if self.model.dropna else 'none').fit()
        mu = mod.params
        # multiply SE by sqrt(N) to turn it into (approx.) SD(Y) on link scale
        sd = (mod.cov_params()[0] * len(mod.mu))**.5

        # modify mu and sd based on means and SDs of slope priors.
        if len(self.model.fixed_terms) > 1 and add_slopes:
            # get order
            index = sum([p['levels'] for p in self.priors.values()], [])
            # get slope prior means and SDs
            means = np.concatenate([p['mu'] \
                for p in self.priors.values()]).ravel()
            means = pd.Series(means, index=index)
            sds = np.concatenate([p['sd'] \
                for p in self.priors.values()]).ravel()
            sds = pd.Series(sds, index=index)
            # add to intercept prior
            mu -= np.dot(means, self.stats['mean_x'][index])
            sd = (sd**2 + np.dot(sds**2, self.stats['mean_x'][index]**2))**.5

        return mu, sd

    def _scale_fixed(self, term, sd_corr):

        # these defaults are only defined for Normal priors
        if term.prior.name != 'Normal':
            return

        mu = []
        sd = []
        for pred in term.data.T:
            mu += [0]
            sd += [self._get_slope_stats(
                exog=self.dm, predictor=pred, sd_corr=sd_corr)]

        # save and set prior
        self.priors.update({term.name: {
            'mu':np.array(mu), 'sd':np.array(sd), 'levels':term.levels,
            }})
        term.prior.update(mu = np.array(mu), sd=np.array(sd))

    def _scale_intercept(self, term, sd_corr):

        # default priors are only defined for Normal priors
        if term.prior.name != 'Normal':
            return

        # get prior mean and SD for fixed intercept
        mu, sd = self._get_intercept_stats()

        # save and set prior
        self.priors.update({term.name: {
            'mu':np.array(mu), 'sd':np.array(sd), 'levels':term.levels,
            }})
        term.prior.update(mu=mu, sd=sd)

    def _scale_random(self, term, sd_corr):
        # these default priors are only defined for HalfNormal priors
        if term.prior.args['sd'].name != 'HalfNormal':
            return

        # recreate the corresponding fixed effect data
        fix_data = term.data.sum(axis=1) \
            if not isinstance(term.data, dict) \
            else np.vstack([term.data[x].sum(axis=1) \
            for x in term.data.keys()]).T

        # classify as random intercept or random slope
        term_type = 'intercept' if np.atleast_2d(fix_data.T).T.sum(1).var()==0 \
            else 'fixed'

        # get name of corresponding fixed effect
        fix = re.sub(r'\|.*', r'', term.name).strip() \
            if term_type=='fixed' else 'Intercept'

        # handle case where there IS a corresponding fixed effect
        if fix in self.model.fixed_terms.keys():
            sd = self.priors[fix]['sd']

        # handle case where there IS NOT a corresponding fixed effect
        else: 
            # handle intercepts and slopes separately
            if term_type=='intercept':
                mu, sd = self._get_intercept_stats()
                sd *= sd_corr
            if term_type=='fixed':
                mu = []
                sd = []
                fix_dataframe = pd.DataFrame(fix_data)
                # things break if column names are integers (the default)
                fix_dataframe.rename(
                    columns={c:'_'+str(c) for c in fix_dataframe.columns},
                    inplace=True)
                exog = self.dm.join(fix_dataframe)
                # this will replace self.mle (which is missing predictors)
                full_mod = sm.GLM(endog=self.model.y.data, exog=exog,
                    family=self.model.family.smfamily(),
                    missing='drop' if self.model.dropna else 'none').fit()
                # loop over the columns of fix_data
                ncols = exog.shape[1] - self.dm.shape[1]
                for pred in range(ncols):
                    mu += [0]
                    sd += [self._get_slope_stats(exog=exog,
                        predictor=np.atleast_2d(fix_data.T).T[:,pred],
                        full_mod=full_mod, sd_corr=sd_corr)]

        # set the prior SD.
        # if there are multiple SDs for multiple categories, use mean for all
        term.prior.args['sd'].update(sd=np.array(sd).mean())

    def scale(self):
        # classify all terms
        fixed_intercepts = [t for t in self.model.terms.values()
            if not t.random and t.data.sum(1).var()==0]
        fixed_slopes = [t for t in self.model.terms.values()
            if not t.random and not t.data.sum(1).var()==0]
        random_terms = [t for t in self.model.terms.values() if t.random]

        # arrange them in the order in which they should be initialized
        term_list = fixed_slopes + fixed_intercepts + random_terms
        term_types = ['fixed']*len(fixed_slopes) + \
            ['intercept']*len(fixed_intercepts) + \
            ['random']*len(random_terms)

        # initialize them in order
        for t, term_type in zip(term_list, term_types):

            # only set default priors if no prior defined yet
            if not isinstance(t.prior, Prior):

                # decide scale
                sd_corr = t.prior
                if sd_corr is None:
                    if not self.model.auto_scale:
                        return
                    sd_corr = 'wide'

                # set scale
                if isinstance(sd_corr, string_types):
                    sd_corr = PriorScaler.names[sd_corr]

                # impute default
                t.prior = self.model.default_priors.get(term=term_type)

                # scale it!
                getattr(self, '_scale_%s' % term_type)(t, sd_corr)




from abc import ABCMeta, abstractmethod
from bambi.external.six import string_types
import numpy as np
import warnings
from bambi.results import PyMC3Results, PyMC3ADVIResults
from bambi.priors import Prior
import theano
try:
    import pymc3 as pm
except:
    warnings.warn("PyMC3 could not be imported. You will not be able to use "
                  "PyMC3 as the back-end for your models.")


class BackEnd(object):

    '''
    Base class for BackEnd hierarchy.
    '''
    __metaclass__ = ABCMeta

    @abstractmethod
    def build(self):
        pass

    @abstractmethod
    def run(self):
        pass


class PyMC3BackEnd(BackEnd):

    '''
    PyMC3 model-fitting back-end.
    '''

    # Available link functions
    links = {
        'identity': lambda x: x,
        'logit': theano.tensor.nnet.sigmoid,
        'inverse': theano.tensor.inv,
        'log': theano.tensor.log
    }

    def __init__(self):
        self.reset()

    def reset(self):
        '''
        Reset PyMC3 model and all tracked distributions and parameters.
        '''
        self.model = pm.Model()
        self.mu = None
        self.dists = {}
        self.shared_params = {}

    def _build_dist(self, label, dist, **kwargs):
        ''' Build and return a PyMC3 Distribution. '''
        if isinstance(dist, string_types):
            if not hasattr(pm, dist):
                raise ValueError("The Distribution class '%s' was not "
                                 "found in PyMC3." % dist)
            dist = getattr(pm, dist)
        # Inspect all args in case we have hyperparameters

        def _expand_args(k, v, label):
            if isinstance(v, Prior):
                label = '%s_%s' % (label, k)
                return self._build_dist(label, v.name, **v.args)
            return v

        kwargs = {k: _expand_args(k, v, label) for (k, v) in kwargs.items()}
        return dist(label, **kwargs)

    def build(self, spec, reset=True):
        '''
        Compile the PyMC3 model from an abstract model specification.
        Args:
            spec (Model): A bambi Model instance containing the abstract
                specification of the model to compile.
            reset (bool): if True (default), resets the PyMC3BackEnd instance
                before compiling.
        '''
        if reset:
            self.reset()

        with self.model:

            self.mu = 0.

            for t in spec.terms.values():

                data = t.data
                label = t.name
                dist_name = t.prior.name
                dist_args = t.prior.args

                # Effects w/ hyperparameters (i.e., random effects)
                if isinstance(data, dict):
                    for level, level_data in data.items():
                        n_cols = level_data.shape[1]
                        mu_label = 'u_%s_%s' % (label, level)
                        u = self._build_dist(mu_label, dist_name,
                                             shape=n_cols, **dist_args)
                        self.mu += pm.math.dot(level_data, u)[:, None]
                else:
                    prefix = 'u_' if t.random else 'b_'
                    n_cols = data.shape[1]
                    coef = self._build_dist(prefix + label, dist_name,
                                            shape=n_cols, **dist_args)
                    self.mu += pm.math.dot(data, coef)[:, None]

            y = spec.y.data
            y_prior = spec.family.prior
            link_f = spec.family.link
            if not callable(link_f):
                link_f = self.links[link_f]
            y_prior.args[spec.family.parent] = link_f(self.mu)
            y_prior.args['observed'] = y
            y_like = self._build_dist(
                spec.y.name, y_prior.name, **y_prior.args)

            self.spec = spec

    def run(self, start=None, method='mcmc', init=None, n_init=10000,
            find_map=False, **kwargs):
        '''
        Run the PyMC3 MCMC sampler.
        Args:
            start: Starting parameter values to pass to sampler; see
                pm.sample() documentation for details.
            method: The method to use for fitting the model. By default,
                'mcmc', in which case the PyMC3 sampler will be used.
                Alternatively, 'advi', in which case the model will be fitted
                using  automatic differentiation variational inference as
                implemented in PyMC3.
            init: Initialization method (see PyMC3 sampler documentation).
                In PyMC3, this defaults to 'advi', but we set it to None.
            n_init: Number of initialization iterations if init = 'advi' or
                'nuts'. Default is kind of in PyMC3 for the kinds of models
                we expect to see run with bambi, so we lower it considerably.
            find_map (bool): whether or not to use the maximum a posteriori
                estimate as a starting point; passed directly to PyMC3.
            kwargs (dict): Optional keyword arguments passed onto the sampler.
        Returns: A PyMC3ModelResults instance.
        '''
        if method == 'mcmc':
            samples = kwargs.pop('samples', 1000)
            with self.model:
                if start is None and find_map:
                    start = pm.find_MAP()
                self.trace = pm.sample(samples, start=start, init=init,
                                       n_init=n_init, **kwargs)
            return PyMC3Results(self.spec, self.trace)

        elif method == 'advi':
            with self.model:
                self.advi_params = pm.variational.advi(start, **kwargs)
            return PyMC3ADVIResults(self.spec, self.advi_params)



import pandas as pd
import numpy as np
import pymc3 as pm
from pymc3.model import TransformedRV
from abc import abstractmethod, ABCMeta
import re, warnings


class ModelResults(object):

    '''
    Base class for ModelResults hierarchy.
    Args:
        model (Model): a bambi Model instance specifying the model.
    '''

    __metaclass__ = ABCMeta

    def __init__(self, model):

        self.model = model
        self.terms = list(model.terms.values())
        self.diagnostics = model._diagnostics \
            if hasattr(model, '_diagnostics') else None
        self.n_terms = len(model.terms)

    @abstractmethod
    def plot(self):
        pass

    @abstractmethod
    def summary(self):
        pass


class PyMC3Results(ModelResults):

    '''
    Holds PyMC3 sampler results and provides plotting and summarization tools.
    Args:
        model (Model): a bambi Model instance specifying the model.
        trace (MultiTrace): a PyMC3 MultiTrace object returned by the sampler. 
    '''

    def __init__(self, model, trace):

        self.trace = trace
        self.n_samples = len(trace)

        # here we determine which variables have been internally transformed 
        # (e.g., sd_log). transformed vars are actually 'untransformed' from 
        # the PyMC3 model's perspective, and it's the backtransformed (e.g., sd)
        # variable that is 'transformed'. So the logic of this is to find and
        # remove 'untranformed' variables that have a 'transformed' counterpart,
        # in that 'untransformed' varname is the 'transfornmed' varname plus
        # some suffix (such as '_log' or '_interval')
        rvs = model.backend.model.unobserved_RVs
        trans = set(var.name for var in rvs if isinstance(var, TransformedRV))
        untrans = set(var.name for var in rvs) - trans
        untrans = set(x for x in untrans if not any([t in x for t in trans]))
        self.untransformed_vars = [x for x in trace.varnames \
            if x in trans | untrans]

        super(PyMC3Results, self).__init__(model)


    def _filter_names(self, names, exclude_ranefs=True, hide_transformed=True):
        names = self.untransformed_vars \
            if hide_transformed else self.trace.varnames
        if exclude_ranefs:
            names = [x for x in names
                if x[2:] not in list(self.model.random_terms.keys())]
        return names


    def plot(self, burn_in=0, names=None, annotate=True, exclude_ranefs=False, 
        hide_transformed=True, kind='trace', **kwargs):
        '''
        Plots posterior distributions and sample traces. Basically a wrapper
        for pm.traceplot() plus some niceties, based partly on code from:
        https://pymc-devs.github.io/pymc3/notebooks/GLM-model-selection.html
        Args:
            burn_in (int): Number of initial samples to exclude before
                summary statistics are computed.
            names (list): Optional list of variable names to summarize.
            annotate (bool): If True (default), add lines marking the
                posterior means, write the posterior means next to the
                lines, and add factor level names for fixed factors with
                more than one distribution on the traceplot.
            exclude_ranefs (bool): If True, do not show trace plots for
                individual random effects. Defaults to False.
            hide_transformed (bool): If True (default), do not show trace
                plots for internally transformed variables.
            kind (str): Either 'trace' (default) or 'priors'. If 'priors',
                this just internally calls Model.plot()
        '''
        if kind == 'priors':
            return self.model.plot()

        # if no 'names' specified, filter out unwanted variables
        if names is None:
            names = self._filter_names(names, exclude_ranefs, hide_transformed)

        # compute means for all variables and factors
        if annotate:
            kwargs['lines'] = {param: self.trace[param][burn_in:].mean() \
                for param in names}
            # factors (fixed terms with shape > 1) must be handled separately
            factors = {}
            for fix in self.model.fixed_terms.values():
                if 'b_'+fix.name in names and len(fix.levels)>1:
                    # remove factor from dictionary of lines to plot
                    kwargs['lines'].pop('b_'+fix.name)
                    # add factor and its column means to dictionary of factors
                    factors.update({'b_'+fix.name:
                        {':'.join(re.findall('\[([^]]+)\]', x)):
                         self.trace['b_'+fix.name][burn_in:].mean(0)[i]
                         for i,x in enumerate(fix.levels)}})

        # make the traceplot
        ax = pm.traceplot(self.trace[burn_in:], varnames=names,
            figsize=(12,len(names)*1.5), **kwargs)

        if annotate:
            # add lines and annotation for the factors
            for f in factors.keys():
                for lev in factors[f]:
                    # draw line
                    ax[names.index(f),0].axvline(x=factors[f][lev],
                        color="r", lw=1.5)
                    # write the mean
                    ax[names.index(f),0].annotate(
                        '{:.2f}'.format(factors[f][lev]),
                        xy=(factors[f][lev],0), xycoords='data', xytext=(5,10),
                        textcoords='offset points', rotation=90, va='bottom',
                        fontsize='large', color='#AA0022')
                    # write the factor level name
                    ax[names.index(f),0].annotate(lev,
                        xy=(factors[f][lev],0), xycoords='data', xytext=(-11,5),
                        textcoords='offset points', rotation=90, va='bottom',
                        fontsize='large', color='#AA0022')
            # add lines and annotation for the rest of the variables
            for v in kwargs['lines'].keys():
                ax[names.index(v),0].annotate(
                    '{:.2f}'.format(kwargs['lines'][v]),
                    xy=(kwargs['lines'][v],0), xycoords='data', xytext=(5,10),
                    textcoords='offset points', rotation=90, va='bottom',
                    fontsize='large', color='#AA0022')

        # For binomial models with n_trials = 1 (most common use case),
        # tell user which event is being modeled
        if self.model.family.name=='binomial' and \
            np.max(self.model.y.data) < 1.01:
            event = next(i for i,x in enumerate(self.model.y.data.flatten()) \
                if x>.99)
            warnings.warn('Modeling the probability that {}==\'{}\''.format(
                self.model.y.name,
                str(self.model.data[self.model.y.name][event])))

        return ax

    def summary(self, burn_in=0, exclude_ranefs=True, names=None,
                hide_transformed=True, **kwargs):
        '''
        Summarizes all parameter estimates. Basically a wrapper for
        pm.df_summary() plus some niceties.
        Args:
            burn_in (int): Number of initial samples to exclude before
                summary statistics are computed.
            exclude_ranefs (bool): If True (default), do not print
                summary statistics for individual random effects.
            names (list): Optional list of variable names to summarize.
            hide_transformed (bool): If True (default), do not print
                summary statistics for internally transformed variables.
        '''

        # if no 'names' specified, filter out unwanted variables
        if names is None:
            names = self._filter_names(names, exclude_ranefs, hide_transformed)

        # get the basic DataFrame
        df = pm.df_summary(self.trace[burn_in:], varnames=names, **kwargs)

        # replace the "__\d" suffixes with an informative factor level name
        match = [re.match('^(.*)(?:__)(\d+)?$', x) for x in df.index]
        def replace_with_name(match):
            term = self.model.terms[match.group(1)[2:]]
            # handle fixed effects
            if term in self.model.fixed_terms.values():
                return term.levels[int(match.group(2))]
            # handle random effects
            else:
                return '{}[{}]'.format(term.name,
                    term.levels[int(match.group(2))])
        new = [replace_with_name(x) if x is not None else df.index[i]
            for i,x in enumerate(match)]
        df.set_index([new], inplace=True)

        # For binomial models with n_trials = 1 (most common use case),
        # tell user which event is being modeled
        if self.model.family.name=='binomial' and \
            np.max(self.model.y.data) < 1.01:
            event = next(i for i,x in enumerate(self.model.y.data.flatten()) \
                if x>.99)
            warnings.warn('Modeling the probability that {}==\'{}\''.format(
                self.model.y.name,
                str(self.model.data[self.model.y.name][event])))

        return df

    def get_trace(self, burn_in=0, names=None, exclude_ranefs=True,
        hide_transformed=True):
        '''
        Returns the MCMC samples in a nice, neat DataFrame.
        Args:
            burn_in (int): Number of initial samples to exclude from
                each chain before returning the trace DataFrame.
            names (list): Optional list of variable names to get samples for.
            exclude_ranefs (bool): If True (default), do not return samples
                for individual random effects.
            hide_transformed (bool): If True (default), do not return
            samples for internally transformed variables.
        '''
        # if no 'names' specified, filter out unwanted variables
        if names is None:
            names = self._filter_names(names, exclude_ranefs, hide_transformed)

        # helper function to label the trace DataFrame columns appropriately
        def get_cols(var):
            # handle terms with a single level
            if len(self.trace[var].shape)==1 or self.trace[var].shape[1]==1:
                return [var]
            else:
                # handle fixed terms with multiple levels
                # (slice off the 'b_' or 'u_')
                if var[2:] in self.model.fixed_terms.keys():
                    return self.model.terms[var[2:]].levels
                # handle random terms with multiple levels
                else:
                    return ['{}[{}]'.format(var[2:], x)
                            for x in self.model.terms[var[2:]].levels]

        # construct the trace DataFrame
        trace_df = pd.concat([pd.DataFrame(self.trace[burn_in:][x],
                                           columns=get_cols(x))
                              for x in names], axis=1)

        return trace_df


class PyMC3ADVIResults(ModelResults):
    '''
    Holds PyMC3 ADVI results and provides plotting and summarization tools.
    Args:
        model (Model): a bambi Model instance specifying the model.
        params (MultiTrace): ADVI parameters returned by PyMC3.
    '''
    def __init__(self, model, params):

        self.means = params['means']
        self.sds = params['stds']
        self.elbo_vals = params['elbo_vals']
        super(PyMC3ADVIResults, self).__init__(model)



import sys
from setuptools import setup, find_packages
from os.path import join, dirname

__version__ = '0.0.5'

if len(set(('test', 'easy_install')).intersection(sys.argv)) > 0:
    import setuptools

requirements = ['numpy', 'pandas', 'patsy', 'pymc3', 'statsmodels']

tests_require = []
extra_setuptools_args = {}
if 'setuptools' in sys.modules:
    tests_require.append('nose')
    extra_setuptools_args = dict(
        test_suite='nose.collector',
        extras_require=dict(
            test='nose>=0.10.1')
    )

setup(
    name="bambi",
    version=__version__,
    description="BAyesian Model Building Interface in Python",
    url='http://github.com/bambinos/bambi',
    download_url='https://github.com/bambinos/bambi/archive/%s.tar.gz' % __version__,
    install_requires=requirements,
    maintainer='Tal Yarkoni',
    maintainer_email='tyarkoni@gmail.com',
    packages=find_packages(exclude=['tests', 'test_*']),
    package_data={'bambi': ["config/*"]},
    tests_require=tests_require,
    license='MIT',
    **extra_setuptools_args
)



def listify(obj):
    ''' Wraps all non-list or tuple objects in a list; provides a simple
    way to accept flexible arguments. '''
    if obj is None:
        return []
    else:
        return obj if isinstance(obj, (list, tuple, type(None))) else [obj]
